# import os
import collections
import sys
from typing import Any

import click

from asmrmanager.cli.dl import dl
from asmrmanager.cli.file import file
from asmrmanager.cli.hold import hold
from asmrmanager.cli.info import info
from asmrmanager.cli.pl import pl
from asmrmanager.cli.play import play
from asmrmanager.cli.query import query
from asmrmanager.cli.review import review
from asmrmanager.cli.sql import sql
from asmrmanager.cli.view import view
from asmrmanager.cli.which import which
from asmrmanager.logger import logger

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
sys.stdout.reconfigure(encoding="utf-8")  # type: ignore
sys.stderr.reconfigure(encoding="utf-8")  # type: ignore


# TODO dry run，clear zip file
# TODO 全局search过滤条件
# TODO query web local local 条件可以简单一些
# TODO 测试各种功能
# catch oserror 然后提醒重试
# view时如果目标文件存在则创建不同后缀名字
# TODO 自动移动字幕文件
class OrderedGroup(click.Group):
    def __init__(self, name=None, commands=None, **attrs):
        super(OrderedGroup, self).__init__(name, commands, **attrs)
        #: the registered subcommands by their exported names.
        self.commands: Any = commands or collections.OrderedDict()

    def list_commands(self, _):
        return self.commands


@click.group(context_settings=CONTEXT_SETTINGS, cls=OrderedGroup)
def main():
    logger.info(f'Run program with: {" ".join(sys.argv[1:])}')


try:
    from trogon import tui  # type: ignore

    main = tui()(main)  # type: ignore
except ImportError:
    pass

main.add_command(dl)
main.add_command(info)
main.add_command(play)
main.add_command(review)
main.add_command(sql)
main.add_command(view)
main.add_command(hold)
main.add_command(query)
main.add_command(which)
main.add_command(file)
main.add_command(pl)

if __name__ == "__main__":
    main()



class DownloadParams:
    def __init__(self, force: bool, replace: bool, filter: bool):
        self.force = force
        self.replace = replace
        self.filter = filter

    @property
    def params(self):
        return {
            "force": self.force,
            "replace": self.replace,
            "filter": self.filter,
        }

    def __str__(self):
        return str(self.params)



from typing import Literal

import click

from asmrmanager.cli.core import fm, id2rj, rj_argument
from asmrmanager.common.rj_parse import RJID
from asmrmanager.filemanager.exceptions import (
    DstItemAlreadyExistsException,
    SrcNotExistsException,
)
from asmrmanager.logger import logger


@click.group(help="some operation about view_path")
def view():
    pass


@click.command()
@rj_argument
@click.option(
    "--mode",
    "-m",
    type=click.Choice(["link", "zip", "adb", "copy"]),
    default="zip",
    show_default=True,
)
def add(rj_id: RJID, mode: Literal["link", "zip", "adb", "copy"]):
    """add an ASMR to view path (use zip by default)"""
    from asmrmanager.cli.core import fm
    from asmrmanager.filemanager.utils import folder_chooser

    src = fm.get_path(rj_id)
    if src is None:
        raise SrcNotExistsException

    rj_name = id2rj(rj_id)
    dst = fm.view_path / rj_name
    if dst.exists():
        i = 1
        while True:
            dst = fm.view_path / f"{rj_name}.{i}"
            if not dst.exists():
                break
            i += 1
        logger.warning('"%s" already exists, use "%s" instead', rj_name, dst)
        # raise DstItemAlreadyExistsException

    src = folder_chooser(src)

    match mode:
        case "zip":
            fm.zip_file(src, dst)
        case "link":
            fm.link(src, dst)
        case "copy":
            fm._copy(src, dst, depth=1)
        case "adb":
            raise NotImplementedError


@click.command("list")
def list_():
    """list all files in view path"""
    rjs = fm.list_("view")
    print(*rjs, sep="\n")


@click.command()
@rj_argument
def remove(rj_id: RJID):
    """remove a file link in view path"""
    fm.remove_view(rj_id)


view.add_command(add)
view.add_command(list_)
view.add_command(remove)



from typing import Iterable, Tuple

import click

from asmrmanager.cli.core import (
    browse_param_options,
    create_database,
    create_downloader_and_database,
    download_param_options,
    fm,
    interval_preprocess_cb,
    multi_rj_argument,
)
from asmrmanager.common.browse_params import BrowseParams
from asmrmanager.common.download_params import DownloadParams
from asmrmanager.common.rj_parse import RJID, id2rj
from asmrmanager.logger import logger


@click.group(help="download ASMR")
def dl():
    pass


@click.command()
@multi_rj_argument
@download_param_options
def get(rj_ids: Iterable[RJID], download_params: DownloadParams):
    """get ASMR by RJ ids"""
    if not rj_ids:
        logger.error("You must give at least one RJ id!")
        return
    spider, db = create_downloader_and_database(download_params)
    spider.run(spider.get(rj_ids))
    db.commit()


@click.command()
@multi_rj_argument
def update(rj_ids: Iterable[RJID]):
    """update metadata, including recover file and description file"""
    if not rj_ids:
        logger.error("You must give at least one RJ id!")
        return
    spider, db = create_downloader_and_database(
        DownloadParams(False, False, True)
    )
    spider.run(spider.update(rj_ids))
    db.commit()


@click.command()
@multi_rj_argument
def check(rj_ids: Iterable[RJID]):
    """check for existence of the file and its store field"""
    db = create_database()
    dl_queue = []
    for rj_id in rj_ids:
        rj = id2rj(rj_id)

        asmr = db.check_exists(rj_id)
        if asmr is None:
            logger.warning(
                f"Not found In Database: {rj}, please manually download it"
            )
            continue

        src = fm.get_location(rj_id)
        if src is None:
            logger.warning(f"Not found: {rj}, add to download")
            dl_queue.append(rj)
            continue

        if src == "download" and asmr.stored:
            logger.info(f"Already stored: {rj}, move to storage path")
            fm.store(rj_id)
            continue

        if src == "storage" and not asmr.stored:
            logger.info(f"Already in storage: {rj}, update database")
            asmr.stored = True
            continue

        logger.info(f"No need to update {rj}")
    db.commit()
    print("Update succesfully.")
    if dl_queue:
        print("Please check the item to download and get theme manually:")
        print(" ".join(dl_queue))


@click.command()
@click.argument("text", type=str, default="")
@click.option(
    "--tags", "-t", type=str, multiple=True, help="tags to include[multiple]"
)
@click.option(
    "--no-tags",
    "-nt",
    type=str,
    multiple=True,
    help="tags to exclude[multiple]",
)
@click.option(
    "--vas",
    "-v",
    type=str,
    multiple=True,
    help="voice actor(cv) to include[multiple]",
)
@click.option(
    "--no-vas",
    "-nv",
    type=str,
    multiple=True,
    help="voice actor(cv) to exclude[multiple]",
)
@click.option(
    "--circle", "-c", type=str, default=None, help="circle(社团) to include"
)
@click.option(
    "--no-circle",
    "-nc",
    type=str,
    multiple=True,
    help="circle(社团) to exclude[multiple]",
)
@click.option(
    "--rate", "-r", help="rating interval", callback=interval_preprocess_cb
)
@click.option(
    "--sell", "-s", help="selling interval", callback=interval_preprocess_cb
)
@click.option(
    "--price", "-pr", help="pirce interval", callback=interval_preprocess_cb
)
@click.option(
    "all_",
    "--all/--select",
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
    help="download all RJs",
)
@browse_param_options
@download_param_options
def search(
    text: str,
    tags: Tuple[str],
    vas: Tuple[str],
    circle: str | None,
    no_tags: Tuple[str],
    no_vas: Tuple[str],
    no_circle: Tuple[str],
    rate: Tuple[float | None, float | None],
    sell: Tuple[int | None, int | None],
    price: Tuple[int | None, int | None],
    browse_params: BrowseParams,
    download_params: DownloadParams,
    all_: bool,
):
    """
    search and download ASMR

    the [multiple] options means you can add multiple same option such as:

        --tags tag1 --tags tag2 --no-tags tag3

    for options like --rate, --sell, --price, you should give a interval like:

        --rate 3.9:4.7 --sell 1000: --price :200

    the interval a:b means a <= x < b, if a or b is not given
    i.e. a: or :b, it means no lower or upper limit
    """
    spider, db = create_downloader_and_database(
        download_params=download_params
    )
    spider.run(
        spider.search(
            text,
            tags=tags,
            vas=vas,
            circle=circle,
            no_tags=no_tags,
            no_vas=no_vas,
            no_circle=no_circle,
            rate=rate,
            sell=sell,
            price=price,
            params=browse_params,
            all_=all_,
        )
    )
    db.commit()


dl.add_command(get)
dl.add_command(check)
dl.add_command(search)
dl.add_command(update)



import click

from asmrmanager.cli.core import rj_argument
from asmrmanager.logger import logger


@click.command()
@rj_argument
@click.option("-c", "--comment", type=str, default=None, show_default=True)
def hold(rj_id: int, comment: str | None = None):
    """set the ASMR.hold to true, and add a comment for the reason"""
    logger.info(f"run command hold with rj_id={rj_id} comment={comment}")

    from asmrmanager.cli.core import create_database

    db = create_database()
    db.hold_item(rj_id, comment)
    db.commit()



from sqlalchemy import Boolean, Column, Date, ForeignKey, Integer, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from asmrmanager.database.utils.uuid_sqlite import GUID

Base = declarative_base()


class ASMR(Base):
    __tablename__ = "asmr"
    id = Column(Integer, primary_key=True)
    title = Column(Text)

    circle_name = Column(Text)  # 对应name字段
    tags = relationship("Tag", secondary="asmrs2tags", backref="asmrs")
    vas = relationship("VoiceActor", secondary="asmrs2vas", backref="asmrs")
    nsfw = Column(Boolean)
    release_date = Column(Date)  # 对应release字段
    price = Column(Integer)
    dl_count = Column(Integer)
    has_subtitle = Column(Boolean)

    star = Column(Integer, default=0)
    count = Column(Integer, default=0)
    comment = Column(Text, default="")

    held = Column(Boolean, default=False)
    stored = Column(Boolean, default=False)


class Tag(Base):
    __tablename__ = "tag"
    id = Column(Integer, primary_key=True, nullable=False)
    name = Column(Text)
    jp_name = Column(Text)
    cn_name = Column(Text)
    en_name = Column(Text)

    def __repr__(self):
        return self.name


class ASMRs2Tags(Base):
    __tablename__ = "asmrs2tags"

    asmr_id = Column(Integer, ForeignKey("asmr.id"), primary_key=True)

    tag_id = Column(Integer, ForeignKey("tag.id"), primary_key=True)


class History(Base):
    __tablename__ = "history"
    id = Column(Integer, primary_key=True, autoincrement=True)
    asmr_id = Column(Integer, ForeignKey("asmr.id"))
    asmr = relationship(ASMR, backref="histories")
    date = Column(Date)
    finish = Column(Boolean, default=False)
    # platform = Column(Enum(PLATFORM))
    # domain = Column(Enum(DOMAIN), default=DOMAIN(data.data.domain))


class ASMRs2VAs(Base):
    __tablename__ = "asmrs2vas"
    asmr_id = Column(Integer, ForeignKey("asmr.id"), primary_key=True)
    actor_id = Column(GUID, ForeignKey("voice_actor.id"), primary_key=True)


class VoiceActor(Base):
    __tablename__ = "voice_actor"
    id = Column(GUID, primary_key=True)
    name = Column(Text)

    def __repr__(self):
        return self.name


def bind_engine(engine):
    Base.metadata.create_all(engine)



from dataclasses import dataclass
from typing import Dict, List, Literal

import toml

from asmrmanager import CONFIG_PATH


@dataclass
class Config:
    username: str
    password: str
    proxy: str
    download_path: str
    storage_path: str
    view_path: str
    tag_filter: List[str]
    editor: str
    filename_filters: List["Filter"]
    download_method: Literal["aria2", "idm"]
    aria2_config: "Aria2Config"
    playlist_aliases: Dict[str, str]


@dataclass
class Filter:
    regex: str = ""  # 正则表达式

    # 该规则是包含还是排除，注意只有满足所有的include规则且不满足所有的exclude规则才会被下载
    type: Literal["include", "exclude"] = "exclude"

    range: Literal["file", "directory", "all"] = "all"  # 该规则针对的文件类型
    excat_match: bool = False  # 是否应精确匹配(从头到尾严格匹配)
    ignore_case: bool = True  # 是否忽略大小写


@dataclass
class Aria2Config:
    host: str = "http://localhost"
    port: int = 6800
    secret: str = ""


_config = toml.load(CONFIG_PATH / "config.toml")

_filename_filters: list = _config.get("filename_filters", [])
filename_filters = list(
    map(
        lambda x: Filter(**x) if isinstance(x, dict) else Filter(x),
        _filename_filters,
    )
)
_aria2_config: dict = _config.get("aria2_config", {})
aria2_config = Aria2Config(**_aria2_config)

playlist_aliases: dict = _config.get("playlist_aliases", {})

config = Config(
    username=_config["username"],
    password=_config["password"],
    proxy=_config["proxy"],
    download_path=_config["download_path"],
    storage_path=_config["storage_path"],
    view_path=_config["view_path"],
    tag_filter=_config["tag_filter"],
    editor=_config["editor"],
    filename_filters=filename_filters,
    download_method=_config["download_method"],
    aria2_config=aria2_config,
    playlist_aliases=playlist_aliases,
)



import os
from collections import Counter
from pathlib import Path

import click
import cutie

from asmrmanager.cli.core import create_database, fm, rj_argument
from asmrmanager.common.rj_parse import RJID, id2rj
from asmrmanager.config import config
from asmrmanager.logger import logger


@click.command()
@click.pass_context
@rj_argument
def play(ctx: click.Context, rj_id: RJID):
    """play an asmr in the terminal"""
    from asmrmanager.lrcplayer import lrc_play

    db = create_database()
    asmr = db.check_exists(rj_id)
    if asmr is None:
        logger.error(f"RJ id {rj_id} not found!")
        return
    asmr_rj = id2rj(asmr.id)

    loc = fm.get_location(asmr.id)
    if loc is None:
        logger.error(f"ASMR {asmr_rj} file not found!")
        return
    rj_path = (
        Path(config.download_path) / asmr_rj
        if loc == "download"
        else Path(config.storage_path) / asmr_rj
    )

    choices = []
    suf = [".mp3", ".wav", ".flac"]
    for path, _, files in os.walk(rj_path):
        res = Counter([Path(i).suffix for i in files])
        if not set(res.keys()).intersection(set(suf)):
            continue
        info = " ".join(
            [f"{key}: {val}" for key, val in res.items() if key in suf]
        )
        choices.append((path, info))

    idx: int = cutie.select([f"{x[0]} | [{x[1]}]" for x in choices])
    # it's just ok to use cutie, no need for beaupy
    # res = select(
    #     choices, preprocessor=lambda x: f'{x[0]} | [{x[1]}]',
    # return_index = True
    # )
    path = Path(choices[idx][0])

    ctx.invoke(lrc_play, path=path)



import re
from typing import Literal

from asmrmanager.config import config


def name_should_download(name: str, type_: Literal["directory", "file"]):
    for filter_ in config.filename_filters:
        range_check = filter_.range == "all" or filter_.range == type_
        if not range_check:
            continue

        flag = re.IGNORECASE if filter_.ignore_case else 0

        match_method = re.fullmatch if filter_.excat_match else re.search

        regex_match = bool(match_method(filter_.regex, name, flag))

        if regex_match ^ (filter_.type == "include"):
            return False

    return True


if __name__ == "__main__":
    print(name_should_download("効果音なしver", type_="directory"))



def print_table(titles, rows, raw=False):
    from rich.console import Console
    from rich.table import Table

    if not raw:
        console = Console()
        table = Table(show_header=True, header_style="bold magenta")
        for k in titles:
            table.add_column(k)
        for row in rows:
            table.add_row(*(str(i) for i in row))
        console.print(table)

    else:
        import json

        res_json = json.dumps(
            [{k: v for k, v in zip(titles, item)} for item in rows],
            ensure_ascii=False,
        )
        import click

        click.echo(res_json)



from pprint import pprint

import click

from asmrmanager.cli.core import (
    create_database,
    create_downloader_and_database,
    rj_argument,
)
from asmrmanager.common.rj_parse import RJID
from asmrmanager.database.orm_type import ASMRInstance
from asmrmanager.logger import logger


def print_asmr_info(asmr: ASMRInstance):
    res: dict = asmr.__dict__.copy()
    res.pop("_sa_instance_state")
    pprint(res)
    print(" tags:", asmr.tags)
    print(" CVs:", asmr.vas)


def info_from_web(rj_id: int):
    downloader, db = create_downloader_and_database()
    (rj_info,) = downloader.run(downloader.downloader.get_voice_info(rj_id))

    res = db.parse_info(rj_info)
    print_asmr_info(res)


@click.command()
@click.option(
    "--rand",
    "-r",
    is_flag=True,
    default=False,
    show_default=True,
    help="get a random info in the database",
)
@rj_argument
def info(rj_id: RJID, rand: bool):
    """show info of the ASMR by id"""
    db = create_database()

    v_info = db.func.get_info(rj_id, rand=rand)
    if v_info is None:
        logger.info(f"RJ{rj_id} not exists, try getting from web")
        info_from_web(rj_id)
        return

    print_asmr_info(v_info)



from datetime import datetime, timedelta
from pathlib import Path

import click


def parse_time(time_str):
    dt = datetime.strptime(time_str.strip(), "%H:%M:%S.%f")
    return dt


def format_time(time):
    return time.strftime("%M:%S.%f")[:8]


DEFAULT_THRESHOLD = timedelta(seconds=2)


def vtt2lrc(vtt, header=True, threshold=DEFAULT_THRESHOLD):
    lrc = ""

    if header:
        lrc += "[re:vtt2lrc]\n"

    last_end = parse_time("23:59:59.99")  # Insanely big value

    for chunk in vtt.split("\n\n")[1:]:
        if not chunk:
            continue

        res = chunk.strip().split("\n", 1)
        if res[0].isdigit():
            time, text = res[1].split("\n", 1)
        else:
            time, text = res

        begin, end = map(parse_time, time.split("-->"))

        if begin - last_end > threshold:
            lrc += f"[{format_time(last_end)}]\n"

        lrc += f"[{format_time(begin)}] {text}\n"

        last_end = end

    lrc += f"[{format_time(last_end)}]\n"

    return lrc


@click.command()
@click.argument(
    "vtt", type=click.Path(exists=True, dir_okay=False, path_type=Path)
)
def main(vtt: Path):
    content = vtt.read_text(encoding="utf-8")
    lrc = vtt2lrc(content)
    with open(
        vtt.with_suffix("").with_suffix(".lrc"), "w", encoding="utf-8"
    ) as f:
        f.write(lrc)



import os
from subprocess import run

import click

from asmrmanager.cli.core import fm
from asmrmanager.common.output import print_table
from asmrmanager.config import config
from asmrmanager.logger import logger


@click.command()
@click.argument("sql_name", type=str)
@click.option(
    "--save/--no-save",
    "-s/-ns",
    is_flag=True,
    default=True,
    help="should your change to the file be saved",
)
def sql(sql_name: str, save: bool):
    """
    execute a sql statement by sql file name in `sqls` directory
    and print the results on your terminal
    """
    sql_path = fm.DATA_PATH / "sqls" / sql_name
    sql_path = sql_path.with_suffix(".sql")
    if not sql_path.exists():
        logger.error(
            f"cannot locate sql {sql_name}, "
            "check ./sqls for possible sql scripts"
        )
        return

    fm.CACHE_PATH.mkdir(parents=True, exist_ok=True)
    temp_file_path = fm.CACHE_PATH / "tempfile.sql"
    temp_file_path.write_text(
        sql_path.read_text(encoding="utf8"), encoding="utf8"
    )
    run(f'{config.editor} "{temp_file_path}"')
    # db_path = temp_file_path.with_name('data.db')
    # print(db_path, temp_file_path)
    # run(
    #     [
    #         'litecli',
    #         str(db_path),
    #         '-e',
    #         temp_file_path.read_text(encoding='utf8'),
    #     ]
    # )
    from asmrmanager.cli.core import create_database

    db = create_database()
    res = db.execute(temp_file_path.read_text(encoding="utf8"))
    rows = res.fetchall()
    titles = res.keys()
    print_table(titles, rows)

    if save:
        sql_path.write_text(
            temp_file_path.read_text(encoding="utf8"), encoding="utf8"
        )
    os.remove(temp_file_path)



import click

from asmrmanager.cli.core import create_database


@click.command()
@click.argument("keyword", type=str)
@click.option("--limit", "-l", type=int, default=100, show_default=True)
@click.option(
    "--raw",
    "-r",
    is_flag=True,
    default=False,
    show_default=True,
    help="output raw json",
)
def query(keyword: str, limit: int, raw: bool):
    """
    simple keyword based query,
    it will match the input with name, circle_name and tag field.
    if you want to use more complex queries, please use `asmr sql` instead.

    set limit to 0 if you want to get all results.
    """
    from asmrmanager.common.output import print_table
    from asmrmanager.database.database import ASMR, ASMRs2Tags, Tag

    db = create_database()
    res_query = (
        db.query(
            ASMR.id,
            ASMR.title,
            ASMR.circle_name,
            ASMR.nsfw,
            ASMR.has_subtitle,
            ASMR.count,
            ASMR.star,
        )
        .join(ASMRs2Tags, ASMRs2Tags.asmr_id == ASMR.id)
        .join(Tag, Tag.id == ASMRs2Tags.tag_id)
        .filter(
            ASMR.circle_name.contains(keyword)
            | ASMR.title.contains(keyword)
            | Tag.name.contains(keyword)
        )
        .group_by(ASMR.id)
    )

    assert limit > 0
    if limit == 0:
        res = res_query.all()
    else:
        res = res_query.limit(limit).all()

    titles = [
        "id",
        "title",
        "circle_name",
        "nsfw",
        "subtitle",
        "count",
        "star",
    ]

    print_table(titles=titles, rows=res, raw=raw)



import os

import click

from asmrmanager.cli.core import rj_argument
from asmrmanager.common.rj_parse import RJID
from asmrmanager.logger import logger


@click.command()
@rj_argument
@click.option("--show", is_flag=True, default=False, show_default=True)
def which(rj_id: RJID, show: bool):
    """Get the path from the given rj_id"""
    from asmrmanager.cli.core import fm

    path = str(fm.get_path(rj_id))

    if not show:
        click.echo(path)
    else:
        import sys

        if sys.platform == "win32":
            os.system(f'explorer.exe "{path}"')
        else:
            logger.error(
                "option --show is for windows only, if you are using *nix, use"
                " `cd $(asmr which <RJID>)`"
            )
            return



from typing import Iterable, List

from .types import RJID, RJName


def rj2id(rj: str) -> RJID | None:
    try:
        if rj.upper().startswith("RJ"):
            return RJID(int(rj[2:]))
        else:
            return RJID(int(rj))
    except ValueError:
        return None


def rjs2ids(rjs: Iterable[str]) -> Iterable[RJID | None]:
    return [rj2id(rj) for rj in rjs]


def rj_name2id(rj_name: RJName) -> RJID:
    return RJID(int(rj_name[2:]))


def rj_names2ids(rjs: Iterable[RJName]) -> Iterable[RJID]:
    return [rj_name2id(rj) for rj in rjs]


def id2rj(rj_id: RJID) -> RJName:
    rj = str(rj_id).zfill(6)
    if len(rj) == 7:
        rj = rj.zfill(8)

    return RJName(f"RJ{rj}")


def ids2rjs(ids: Iterable[RJID]) -> List[RJName]:
    return [id2rj(id_) for id_ in ids]


if __name__ == "__main__":
    rj = RJName("RJ123456")
    print(rj2id(rj))



import contextlib
from dataclasses import dataclass
from typing import List, Tuple

import pylrc

with contextlib.redirect_stdout(None):
    from pygame import mixer

from pathlib import Path

from chardet import detect
from mutagen._file import File as MutagenFile


class MusicPlayer:
    STEP_TIME = 15

    def __init__(self, music_path: Path) -> None:
        self.music_path = music_path
        self.title = music_path.name
        if mixer.get_init() is None:
            mixer.init()
        self.player = mixer.music
        self.prev_time = 0
        self.prev_pos = 0

        self.total_time = int(
            MutagenFile(music_path).info.length * 1000  # type:ignore
        )

    def is_playing(self):
        return self.player.get_busy()

    def play(self):
        self.prev_time = 0
        self.prev_pos = 0
        self.player.load(self.music_path)
        self.player.play()

    def stop(self):
        self.player.stop()

    def _get_play_time(self):
        return self.player.get_pos()

    def get_time_pass(self):
        return self._get_play_time() - self.prev_time

    def get_pos(self):
        return self.prev_pos + self.get_time_pass()

    def pause(self):
        mixer.music.pause()

    def unpause(self):
        mixer.music.unpause()

    def _set_pos(self, pos: int):
        # 对于wav类型的文件，set_pos可能会抛出错误，如果postion超出范围了的话
        try:
            self.player.set_pos(pos / 1000)
        except Exception:
            return self._set_pos(pos - 500)
        self.prev_pos = pos
        self.prev_time = self._get_play_time()

    def forward(self, t: float | None = None):
        t_ = t if t is not None else self.STEP_TIME
        dst = self.get_pos() + int(t_ * 1000)
        self._set_pos(dst)

    def backward(self, t: float | None = None):
        t_ = t if t is not None else self.STEP_TIME
        dst = self.get_pos() - int(t_ * 1000)
        self._set_pos(dst)

    def get_total_percent(self) -> int:
        return min(int(self.get_pos() / self.total_time * 100), 100)


class LyricsParser:
    @dataclass
    class LyricsData:
        progress: int
        lyrics: List[str]

    def __init__(self, lrc_path: Path, total_time: int) -> None:
        content = lrc_path.read_bytes()
        encoding = detect(content)["encoding"]
        if encoding is None:
            raise EncodingWarning("Unrecognizable encoding")
        encoding = "gbk" if encoding.upper() == "GB2312" else encoding
        self.lrc = self.parse_lrc(content.decode(encoding))

        self.total_time = total_time

    @staticmethod
    def parse_lrc(lrc_str: str) -> List[Tuple[int, str]]:
        lrc = pylrc.parse(lrc_str)
        res = [
            (int(line.time * 1000), line.text)
            for line in lrc
            if line.text.strip()
        ]
        return sorted(res, key=lambda x: x[0])

    def get_next_index(self, t: int):
        for i, line in enumerate(self.lrc):
            if t < line[0]:
                return i
        else:
            return -1

    def get_lrc_data_by_time(self, t: int):
        def get_by_index(idx: int):
            res = []
            for i in (idx - 2, idx - 1, idx):
                if 0 <= i < len(self.lrc):
                    res.append(self.lrc[i][1])
                else:
                    res.append("")
            return res

        idx = self.get_next_index(t)
        if idx == -1:
            t1 = t - self.lrc[-1][0]
            t2 = self.total_time - self.lrc[-1][0]

            # 对于wav文件来说，get_pos 的最大值会和 total_time 有些许出入，误差大概几百个毫秒，
            # 一般都是get_pos更大，另外total_time是准确值
            percentage = 0 if t2 == 0 else min(int(t1 / t2 * 100), 100)
            return self.LyricsData(percentage, get_by_index(len(self.lrc)))

        if idx == 0:
            return self.LyricsData(
                int(t / self.lrc[0][0] * 100) if self.lrc[0][0] != 0 else 0,
                get_by_index(idx),
            )

        if self.lrc[idx][0] - self.lrc[idx - 1][0] != 0:
            progress = int(
                (t - self.lrc[idx - 1][0])
                / (self.lrc[idx][0] - self.lrc[idx - 1][0])
                * 100
            )
        else:
            progress = 0
        return self.LyricsData(progress, get_by_index(idx))


class MusicPlayerWithLyrics(MusicPlayer, LyricsParser):
    def __init__(self, music_path: Path, lrc_path: Path) -> None:
        MusicPlayer.__init__(self, music_path)
        LyricsParser.__init__(self, lrc_path, self.total_time)

    def get_lyrics(self):
        return self.get_lrc_data_by_time(self.get_pos())

    def set_pos_by_index(self, idx: int):
        pos = self.lrc[idx][0]
        self._set_pos(pos)

    def forward_lrc(self):
        idx = self.get_next_index(self.get_pos())
        if idx == -1:
            return
        self.set_pos_by_index(idx)

    def backward_lrc(self):
        idx = self.get_next_index(self.get_pos())
        if idx in (0, 1):
            self._set_pos(0)
            return
        if idx == -1:
            self.set_pos_by_index(-2)
            return
        self.set_pos_by_index(idx - 2)


if __name__ == "__main__":
    p = MusicPlayer(
        Path(r"E:\asmr\RJ339431\1日目 はじめてのさいみん(手マン).flac")
    )
    lrc = LyricsParser(
        Path(r"E:\asmr\RJ339431\1日目 はじめてのさいみん(手マン).lrc"),
        p.total_time,
    )
    pl = MusicPlayerWithLyrics(
        Path(r"E:\asmr\RJ339431\1日目 はじめてのさいみん(手マン).flac"),
        Path(r"E:\asmr\RJ339431\1日目 はじめてのさいみん(手マン).lrc"),
    )



"""
Python IDM Helper by zackmark29
Version v1.0.2 | 2021.11.28
"""

from pathlib import Path
from types import ModuleType
from typing import Any, Optional

from comtypes import client  # type: ignore
from comtypes.automation import VT_EMPTY  # type: ignore


def check_folder(*dirs: Path) -> Path:
    for d in dirs:
        if d.exists():
            return d
    raise NotADirectoryError(
        "Looks like you don't have IDM installed in your system"
    )


def get_module() -> ModuleType:
    idm_folder_64bit = Path(
        r"C:\Program Files (x86)\Internet Download Manager"
    )
    idm_folder_32bit = Path(r"C:\Program Files\Internet Download Manager")

    idm_folder = check_folder(idm_folder_64bit, idm_folder_32bit)

    try:
        # Registry path: Computer\HKEY_CLASSES_ROOT\TypeLib\
        # {ECF21EAB-3AA8-4355-82BE-F777990001DD}
        UUID = "{ECF21EAB-3AA8-4355-82BE-F777990001DD}"
        return client.GetModule([UUID, 1, 0])
    except OSError:
        # if uuid is not exist in the registry use tlb module instead
        tlb = idm_folder / "idmantypeinfo.tlb"
        if not tlb.exists():
            raise FileNotFoundError(
                f"{tlb} is not exist. Try to re-install your idm"
            )

        return client.GetModule(tlb)


class IDMHelper:
    """
    flags
        0: Display/Pop-up confirmation before downloading
        1: Download automatically without any confirmations dialogs
        2: Display confirmation if found duplicate and add only to queue
        3: Add only to queue without any confirmation
    """

    def __init__(
        self,
        url: str,
        output_folder: str,
        output_file_name: str,
        flag: int,
        referer: Optional[str] = None,
        cookies: Optional[str] = None,
        post_data: Optional[str] = None,
        user_name: Optional[str] = None,
        password: Optional[str] = None,
        user_agent: Optional[str] = None,
    ) -> None:
        # common
        self.url = url
        self.flag = flag
        self.output_folder = output_folder
        self.output_filename = output_file_name

        # optionals
        self.referer = referer
        self.cookies = cookies
        self.post_data = post_data
        self.user_name = user_name
        self.password = password
        self.user_agent = user_agent

        self.idm_module = get_module()

    def send_link_to_idm(self) -> int:
        """Simple method with limited parameters"""

        idm: Any = client.CreateObject(
            progid="IDMan.CIDMLinkTransmitter",
            interface=self.idm_module.ICIDMLinkTransmitter,
        )

        res: int = idm.SendLinkToIDM(
            bstrUrl=self.url,
            bstrLocalPath=self.output_folder,
            bstrLocalFileName=self.output_filename,
            lFlags=self.flag,
            bstrReferer=None,
            bstrCookies=None,
            bstrData=None,
            bstrUser=None,
            bstrPassword=None,
        )
        return res

    def send_link_to_idm2(self) -> None:
        """Method with all the parameters"""

        idm: Any = client.CreateObject(
            progid="IDMan.CIDMLinkTransmitter",
            interface=self.idm_module.ICIDMLinkTransmitter2,
        )
        idm.SendLinkToIDM2(
            bstrUrl=self.url,
            bstrReferer=self.referer,
            bstrCookies=self.cookies,
            bstrData=self.post_data,
            bstrUser=self.user_name,
            bstrPassword=self.password,
            bstrLocalPath=self.output_folder,
            bstrLocalFileName=self.output_filename,
            lFlags=self.flag,
            reserved1=self.user_agent if self.user_agent else VT_EMPTY,
            reserved2=VT_EMPTY,
        )


if __name__ == "__main__":
    url = "http://www.internetdownloadmanager.com/idman401.exe"

    output_folder = Path.cwd()
    output_filename = "idman.exe"
    user_agent = None
    flag = 3  # see above the flag information
    idm = IDMHelper(url, str(output_folder), output_filename, flag)
    idm.send_link_to_idm()
    # idm.send_link_to_idm2()



import json
import os

from asmrmanager.database.manage import DataBaseManager


def add2db(start_dir: str):
    db = DataBaseManager()
    for root, dirs, files in os.walk(start_dir):
        for file in files:
            if file.startswith("RJ") and file.endswith(".json"):
                with open(os.path.join(root, file), encoding="utf8") as f:
                    db.add_info(json.load(f))

    db.commit()


if __name__ == "__main__":
    add2db(r"D:\bit_download\asmr")



import json
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, TypeVar

T = TypeVar("T", bound="ASMRDownloadAPI")


from asmrmanager.common.rj_parse import RJID, id2rj
from asmrmanager.config import Aria2Config
from asmrmanager.logger import logger
from asmrmanager.spider.asmrapi import ASMRAPI

try:
    IDMHELPER_EXIST = True
    from .utils.IDMHelper import IDMHelper
except (ImportError, ModuleNotFoundError):
    IDMHELPER_EXIST = False

try:
    ARIA2_EXIST = True
    from .utils.aria2_downloader import Aria2Downloader
except (ImportError, ModuleNotFoundError):
    ARIA2_EXIST = False

from typing import NamedTuple

from asmrmanager.cli.core import fm

# TODO 统一管理固定的参数

FileInfo = NamedTuple(
    "FileInfo", [("path", Path), ("url", str), ("should_download", bool)]
)


class ASMRDownloadAPI(ASMRAPI):
    def __init__(
        self,
        name: str,
        password: str,
        proxy: str,
        json_should_download: Callable[[Dict[str, Any]], bool],
        name_should_download: Callable[
            [str, Literal["directory", "file"]], bool
        ],
        replace=False,
        limit: int = 3,
        download_method: Literal["aria2", "idm"] = "idm",
        aria2_config: Aria2Config | None = None,
    ):
        # self._session: Optional[ClientSession] = None  # for __aenter__
        super().__init__(name, password, proxy, limit)
        self.save_path = fm.download_path
        # self.pop_keys = (
        #             "create_date",
        #             "userRating",
        #             "review_text",
        #             "progress",
        #             "updated_at",
        #             "user_name",
        #             "rate_count_detail",
        #             "rank"
        #     )
        self.json_should_download = json_should_download
        self.name_should_download = name_should_download
        self.replace = replace
        self.download_file = (
            self.download_by_idm
            if download_method == "idm"
            else self.download_by_aria2
        )

        self.aria2_downloader = Aria2Downloader(proxy)
        self.aria2_config = aria2_config
        self.download_method = download_method

    async def download(
        self,
        voice_id: int,
        save_path: Path | None = None,
    ) -> None:
        voice_info = await self.get_voice_info(voice_id)

        should_down = self.json_should_download(voice_info)
        if not should_down:
            logger.info(f"stop download {voice_id}")
            return
        if save_path is None:
            save_path = self.save_path

        voice_path = save_path / id2rj(RJID(voice_id))
        if voice_path.exists():
            logger.warning(f"path {voice_path} already exists.")

        voice_path.mkdir(parents=True, exist_ok=True)
        self.create_info_file(voice_info, voice_path=voice_path)

        tracks = await self.get_voice_tracks(voice_id)
        if isinstance(tracks, dict):
            if error_info := tracks.get("error"):
                logger.error(f"RJ{voice_id} not found, {error_info}")
                return
            else:
                logger.error("Unexpected track type: dict")
                return

        file_list = self.get_file_list(tracks, voice_path)
        self.create_recover_file(file_list, voice_path)
        await self.create_dir_and_download(file_list)

    def create_recover_file(self, file_list: List[FileInfo], voice_path: Path):
        recover = [
            {
                "path": str(file.path.relative_to(voice_path)).replace(
                    "\\", "/"
                ),
                "url": file.url,
                "should_download": file.should_download,
            }
            for file in file_list
        ]
        with open(voice_path / ".recover", "w", encoding="utf-8") as f:
            json.dump(recover, f, ensure_ascii=False, indent=4)

    async def get_voice_info(self, voice_id: int) -> Dict[str, Any]:
        voice_info = await self.get(f"work/{voice_id}")
        assert isinstance(voice_info, dict)
        return voice_info

    async def get_voice_tracks(self, voice_id: int):
        return await self.get(f"tracks/{voice_id}")

    @staticmethod
    async def download_by_idm(
        url: str, save_path: Path, file_name: str
    ) -> bool:
        """the save path + file should not exist,
        and the filename should be legal"""
        assert IDMHELPER_EXIST
        m = IDMHelper(url, str(save_path.absolute()), file_name, 3)
        res = m.send_link_to_idm()
        if res != 0:
            logger.error("IDM returns an error code!")
            return False
        return True

    async def download_by_aria2(
        self, url: str, save_path: Path, file_name: str
    ) -> bool:
        """the save path + file should not exist,
        and the filename should be legal"""

        await self.aria2_downloader.download(url, save_path, file_name)
        return True

    def check_exists(self, download_file_path: Path):
        p = (
            fm.download_path
            if download_file_path.is_relative_to(fm.download_path)
            else fm.storage_path
        )
        rel_path = str(download_file_path.relative_to(p))
        return fm.check_exists(rel_path)

    async def process_download(
        self, url: str, save_path: Path, file_name: str
    ):
        # file_name = file_name.translate(
        #     str.maketrans(r'/\:*?"<>|', "_________")
        # )

        file_path = save_path / file_name
        exist_info = self.check_exists(file_path)
        if exist_info.download and self.replace:
            logger.info(f"replace mode, delete old file {file_path}")
            file_path.unlink()

        if exist_info.download:
            logger.warning(
                f"file {file_path} already exists in download, ignore this"
                " file"
            )
            return

        if exist_info.storage:
            logger.warning(
                f"file {file_path} already exists in storage, ignore this file"
            )
            return

        logger.info(f"Downloading {file_path}")
        if not await self.download_file(url, save_path, file_name):
            logger.error(f"Download {file_path} failed")
            return

    def create_info_file(self, voice_info: Dict[str, Any], voice_path: Path):
        rj_name = id2rj(voice_info["id"])
        # info中有名字信息，理论上应该是一样的，但有可能为空，所以不考虑使用
        # recv_rj_name = voice_info['original_workno']
        json_path = voice_path / f"{rj_name}.json"
        if json_path.exists():
            logger.info(f"Path {json_path} already exists, update it...")
        with open(json_path, "w", encoding="utf-8") as f:
            json.dump(voice_info, f, ensure_ascii=False, indent=4)

    async def create_dir_and_download(self, file_list: List[FileInfo]) -> None:
        for file_info in file_list:
            file_path = file_info.path
            file_path.parent.mkdir(parents=True, exist_ok=True)
            if not file_info.should_download:
                logger.info(f"filter file {file_path}")
                # with open(
                #     file_path.with_suffix(file_path.suffix + ".info"),
                #     "w",
                #     encoding="utf-8",
                # ) as f:
                #     f.write(file["mediaDownloadUrl"])
                continue
            try:
                await self.process_download(
                    file_info.url, file_path.parent, file_path.name
                )
            except ModuleNotFoundError as e:
                logger.critical(
                    f"Module not found: {e}, please read the install part of"
                    " the README.md to install the corresponding module"
                )
                exit(-1)
            except Exception as e:
                logger.error(f"Unknow download error: {e}")
                continue

    def get_file_list(
        self, tracks: List[Dict[str, Any]], voice_path: Path, download=True
    ):
        file_list: List[FileInfo] = []
        folders = [track for track in tracks if track["type"] == "folder"]
        files = [track for track in tracks if track["type"] != "folder"]

        for file in files:
            file_name: str = file["title"].translate(
                str.maketrans(r'/\:*?"<>|', "_________")
            )
            file_path = voice_path / file_name

            if (not download) or (
                not self.name_should_download(file["title"], "file")
            ):
                should_download = False
            else:
                should_download = True

            file_list.append(
                FileInfo(file_path, file["mediaDownloadUrl"], should_download)
            )

        for folder in folders:
            download_ = (
                True
                if self.name_should_download(folder["title"], "directory")
                else False
            )
            title: str = folder["title"].translate(
                str.maketrans(r'/\:*?"<>|', "_________")
            )
            new_path = voice_path / title
            file_list.extend(
                self.get_file_list(
                    folder["children"], new_path, download and download_
                )
            )

        return file_list

    async def __aenter__(self: T) -> T:
        await super().__aenter__()
        if self.download_method == "aria2":
            assert self.aria2_config
            await self.aria2_downloader.create_client(
                self.aria2_config.host,
                self.aria2_config.port,
                self.aria2_config.secret,
            )
        return self

    async def __aexit__(self, *args) -> None:
        await super().__aexit__(*args)
        await self.aria2_downloader.close_client()



from dataclasses import dataclass
from typing import List, Literal


@dataclass
class Config:
    username: str
    password: str
    proxy: str
    download_path: str
    storage_path: str
    view_path: str
    tag_filter: List[str]
    editor: str
    filename_filters: List["Filter"]
    download_method: Literal["aria2", "idm"]
    aria2_config: "Aria2Config"


@dataclass
class Filter:
    regex: str = ""  # 正则表达式

    # 该规则是包含还是排除，注意只有满足所有的include规则且不满足所有的exclude规则才会被下载
    type: Literal["include", "exclude"] = "exclude"

    range: Literal["file", "directory", "all"] = "all"  # 该规则针对的文件类型
    excat_match: bool = False  # 是否应精确匹配(从头到尾严格匹配)
    ignore_case: bool = True  # 是否忽略大小写


@dataclass
class Aria2Config:
    host: str = "http://localhost"
    port: int = 6800
    secret: str = ""


# -----------------------------------------------------
# 以下是配置文件的示例，上方的内容请勿修改
# -----------------------------------------------------


config = Config(
    # 账号
    username="username",
    # 密码
    password="password",
    # 代理
    proxy="http://localhost:7890",
    # 下载文件的临时存放位置
    download_path="/path/to/some/place",
    # 文件最终存储位置
    # 使用review后，如果文件在download_path，将自动移至storage_path
    storage_path="/path/to/some/place",
    # 使用asmr view 的时候会在这个位置创建一个软链接或压缩文件，方便查找。
    # 如不使用 view 方法可以设为空
    view_path="/path/to/some/place",
    # 过滤掉那些你不喜欢的标签
    tag_filter=["tagname1", "tagname2"],
    # 编辑sql用的默认编辑器, 如果需使用powershell脚本例如lvim，
    # 此处可填入 'pwsh -nop -f /path/to/lvim.ps1'
    editor="code --wait",
    # 文件和文件夹的过滤规则
    filename_filters=[
        # 过滤所有的SE(不完全)
        Filter(r"(効果音|SE|ＳＥ|BGM|音效)([な無无×][し]?|cut|切除|カット)"),
        Filter(r"([無无×]|不含|NO[ _]?)(効果音|SE|ＳＥ|BGM|音效)"),
        Filter(r"(声|ボイス|SE)のみ"),
        # 过滤所有的wav
        Filter(r".*\.wav$", range="file"),
        # 过滤包含 "反転" 的wav文件
        Filter(r"反転.*\.wav$", range="file"),
    ],
    # 下载所用的工具
    download_method="idm",
    # 如使用aria2请配置此项，否则请忽略
    aria2_config=Aria2Config(),
)



import math
from datetime import date
from typing import Any, Dict, Sequence, Union, cast

import sqlalchemy.orm
from sqlalchemy import event, text
from sqlalchemy.engine import Engine, ResultProxy
from sqlalchemy.engine.result import Result
from sqlalchemy.orm import Session, sessionmaker

from asmrmanager.database.orm_type import ASMRInstance
from asmrmanager.logger import logger

from .database import ASMR, Tag, VoiceActor, bind_engine
from .engine import get_engine
from .q_func import QFunc


def create_math_functions_on_connect(dbapi_connection, connection_record):
    dbapi_connection.create_function("sin", 1, math.sin)
    dbapi_connection.create_function("cos", 1, math.cos)
    dbapi_connection.create_function("acos", 1, math.acos)
    dbapi_connection.create_function("radians", 1, math.radians)
    dbapi_connection.create_function("log2", 1, math.log2)
    dbapi_connection.create_function("log10", 1, math.log10)


class DataBaseManager:
    def __init__(
        self,
        engine: Union[Engine, None] = None,
        tag_filter: Sequence[str] = tuple(),
    ):
        global create_math_functions_on_connect
        self.engine = engine or get_engine()
        event.listens_for(self.engine, "connect")(
            create_math_functions_on_connect
        )  # add a listener after engine created, before session created
        bind_engine(self.engine)

        self.tag_filter = set(tag_filter)

        self.session: Session = sessionmaker(self.engine)()
        self.func = QFunc(self.session)

    def check_exists(self, rj_id: int) -> Union[ASMRInstance, None]:
        return self.session.query(ASMR).get(rj_id)

    @classmethod
    def parse_info(cls, info: Dict[str, Any]) -> ASMRInstance:
        asmr = ASMR(
            id=info["id"],
            title=info["title"],
            circle_name=info["name"],
            nsfw=info["nsfw"],
            release_date=date.fromisoformat(info["release"]),
            price=info["price"],
            dl_count=info["dl_count"],
            has_subtitle=info["has_subtitle"],
        )

        for actor_info in info["vas"]:
            actor = VoiceActor(**actor_info)
            asmr.vas.append(actor)

        for tag_info in info["tags"]:
            if not tag_info.get("id"):
                continue
            tag = Tag(id=tag_info["id"], name=tag_info["name"])
            if tag_info["i18n"]:
                tag.cn_name = tag_info["i18n"]["zh-cn"]["name"]
                tag.jp_name = tag_info["i18n"]["ja-jp"]["name"]
                tag.en_name = tag_info["i18n"]["en-us"]["name"]
            else:
                assert tag.id == 10000
            asmr.tags.append(tag)
        return asmr

    def add_info(self, info: Dict[str, Any]) -> bool:
        """
        add/update info to database and check
        if it has tag in the filter or not,
        return True if should download
        """

        asmr = self.parse_info(info)

        self.session.merge(asmr)

        # check for tag filter
        tags = [t["name"] for t in info["tags"]]
        if self.tag_filter.intersection(tags):
            logger.info(f"ignore {asmr.id} since it has tags: {tags}")
            return False
        return True

    def update_review(
        self, rj_id: int, star: int, comment: str, update_stored: bool = False
    ):
        if not (asmr := self.check_exists(rj_id)):
            logger.error("Incorrect RJ ID, no item in database!")
            exit(-1)

        asmr.count += 1

        if star is not None:
            if (not isinstance(star, int)) or star < 1 or star > 5:
                logger.error("Your star should be a integer between 1 and 5")
                exit(-1)
            asmr.star = star

        if comment is not None:
            comment = f"{date.today()}: {comment}\n"
            asmr.comment += comment

        if update_stored:
            asmr.stored = True

    def hold_item(self, rj_id: int, comment: str | None):
        if not (asmr := self.check_exists(rj_id)):
            logger.error("Incorrect RJ ID, no item in database!")
            return

        if asmr.held:
            logger.warning(f"ASMR id={rj_id} has already been held!")

        asmr.held = True
        if comment is not None:
            comment = f"{date.today()}: {comment}\n"
            asmr.comment += comment

    def execute(self, sql: str) -> ResultProxy | Result:
        return self.session.execute(text(sql))

    def query(self, *args, **kwargs) -> sqlalchemy.orm.Query:
        return cast(sqlalchemy.orm.Query, self.session.query(*args, **kwargs))

    def commit(self):
        self.session.commit()
        logger.info("successfully committed")



import zipfile
from pathlib import Path

from asmrmanager.logger import logger

from .utils import folder_chooser


def zip_chosen_folder(src: Path, dst: Path):
    assert src.is_dir() and dst.exists() is False
    folder = folder_chooser(src)
    logger.info("start zipping...")
    files = [f for f in folder.iterdir() if not f.is_dir()]
    with zipfile.ZipFile(dst, "w", zipfile.ZIP_DEFLATED) as zf:
        for file in files:
            zf.write(file, arcname=file.name)



import uuid
from collections.abc import Iterable
from typing import Any, List, Tuple

from asmrmanager.common.rj_parse import RJID
from asmrmanager.common.types import PRIVACY, PlayListItem

from .asmrapi import ASMRAPI


class ASMRPlayListAPI(ASMRAPI):
    def __init__(
        self, name: str, password: str, proxy: str, limit: int = 3
    ) -> None:
        super().__init__(name, password, proxy, limit)

    async def get_playlists(
        self, page: int = 1, page_size: int = 12, filter_by: str = "all"
    ) -> Tuple[List[PlayListItem], int]:
        """return a list of playlists and the total"""
        resp = await super().get_playlists(page, page_size, filter_by)
        return (
            self.process_playlists(resp["playlists"]),
            resp["pagination"]["totalCount"],
        )

    async def create_playlist(
        self,
        name: str,
        desc: str | None = None,
        privacy: PRIVACY = PRIVACY.PRIVATE,
    ):
        return await super().create_playlist(name, desc, privacy.value)

    async def add_works_to_playlist(
        self, rj_ids: Iterable[RJID], pl_id: uuid.UUID
    ):
        return await super().add_works_to_playlist(list(rj_ids), str(pl_id))

    async def delete_playlist(self, pl_id: uuid.UUID):
        return await super().delete_playlist(str(pl_id))

    def process_playlists(self, playlists: List[Any]) -> List[PlayListItem]:
        res = []
        for item in playlists:
            res.append(
                PlayListItem(
                    id=uuid.UUID(item["id"]),
                    name=item["name"],
                    privacy=PRIVACY(item["privacy"]),
                    desc=item["description"],
                    works_count=item["works_count"],
                    latest_work_id=RJID(item["latestWorkID"]),
                )
            )
        return res



import asyncio
import uuid
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Tuple,
    TypeVar,
)

import cutie

from asmrmanager.cli.core import fm
from asmrmanager.common.browse_params import BrowseParams
from asmrmanager.common.rj_parse import RJID, id2rj
from asmrmanager.config import Aria2Config
from asmrmanager.logger import logger
from asmrmanager.spider.asmrapi import ASMRAPI
from asmrmanager.spider.playlist import ASMRPlayListAPI

from .downloader import ASMRDownloadAPI

T = TypeVar("T", bound=Any)


class AsyncManager:
    def __init__(self, api: ASMRAPI) -> None:
        self.api = api

    def run(self, *tasks: Awaitable[T]) -> List[T]:
        async def _run():
            async with self.api:
                return await asyncio.gather(*tasks)

        loop = asyncio.get_event_loop()
        return loop.run_until_complete(_run())


class ASMRDownloadManager(AsyncManager):
    def __init__(
        self,
        name: str,
        password: str,
        proxy: str,
        id_should_download: Callable[[RJID], bool] | None = None,
        json_should_download: Callable[[Dict[str, Any]], bool] | None = None,
        name_should_download: Callable[
            [str, Literal["directory", "file"]], bool
        ]
        | None = None,
        replace=False,
        download_method: Literal["aria2", "idm"] = "idm",
        aria2_config: Aria2Config | None = None,
    ):
        self.downloader = ASMRDownloadAPI(
            name=name,
            password=password,
            proxy=proxy,
            name_should_download=name_should_download or (lambda *_: True),
            json_should_download=json_should_download or (lambda _: True),
            replace=replace,
            download_method=download_method,
            aria2_config=aria2_config,
        )
        super().__init__(self.downloader)
        self.id_should_download = id_should_download or (lambda _: True)

    async def get(self, ids: Iterable[RJID]):
        tasks = []
        for arg in ids:
            if not self.id_should_download(arg):
                logger.info(f"RJ{arg} already exists.")
                continue
            tasks.append(self.downloader.download(arg))
        await asyncio.gather(*tasks)

    async def search(
        self,
        text: str,
        tags: Tuple[str],
        vas: Tuple[str],
        circle: str | None,
        no_tags: Tuple[str],
        no_vas: Tuple[str],
        no_circle: Tuple[str],
        rate: Tuple[float | None, float | None],
        sell: Tuple[int | None, int | None],
        price: Tuple[int | None, int | None],
        params: BrowseParams,
        all_: bool,
    ):
        filters = []

        filters += [f"$tag:{t}$" for t in tags]
        filters += [f"$-tag:{nt}$" for nt in no_tags]

        filters += [f"$va:{va}$" for va in vas]
        filters += [f"$-va:{nva}$" for nva in no_vas]

        filters += [f"$-circle:{nc}$" for nc in no_circle]

        for name, value in (("rate", rate), ("sell", sell), ("price", price)):
            if value[0] is not None:
                filters.append(f"${name}:{value[0]}$")
            if value[1] is not None:
                filters.append(f"$-{name}:{value[1]}$")

        if circle:
            filters.append(f"$circle:{circle}$")
        if text:
            filters.append(text)

        if filters:
            logger.info(f"searching with {filters} {params}")
            search_result = await self.downloader.get_search_result(
                " ".join(filters).replace("/", "%2F"), params=params.params
            )
        else:
            logger.info(f"list works with {params}")
            search_result = await self.downloader.list(params=params.params)
        ids = [work["id"] for work in search_result["works"]]

        if all_:
            await self.get(ids)
            return

        # select RJs
        titles = [work["title"] for work in search_result["works"]]
        indexes = cutie.select_multiple(
            [f"{id2rj(id_)} | {title}" for id_, title in zip(ids, titles)],
        )
        if not indexes:
            logger.error("Nothing was selected.")
            return

        await self.get([ids[i] for i in indexes])

    async def tag(self, tag_name: str, params: BrowseParams):
        """tag 和 va 一样，都是调用了特殊的search方法"""
        tag_res = await self.downloader.tag(tag_name, params=params.params)
        ids = [work["id"] for work in tag_res["works"]]
        await self.get(ids)

    async def va(self, va_name: str, params: BrowseParams):
        """tag 和 va 一样，都是调用了特殊的search方法"""
        va_res = await self.downloader.tag(va_name, params=params.params)
        ids = [work["id"] for work in va_res["works"]]
        await self.get(ids)

    async def update(self, ids: Iterable[RJID]):
        async def update_one(rj_id_: RJID):
            voice_info = await self.downloader.get_voice_info(rj_id_)

            # should_down = self.spider.json_should_download(voice_info)
            # if not should_down:
            #     logger.info(f"stop download {rj_id_}")
            #     return
            save_path = fm.storage_path

            voice_path = save_path / id2rj(rj_id_)
            if not voice_path.exists():
                logger.warning(
                    f"There are such file in your storage path for RJ{rj_id_}"
                )

            voice_path.mkdir(parents=True, exist_ok=True)
            self.downloader.create_info_file(voice_info, voice_path)

            tracks = await self.downloader.get_voice_tracks(rj_id_)
            if isinstance(tracks, dict):
                if error_info := tracks.get("error"):
                    logger.error(f"RJ{rj_id_} not found, {error_info}")
                    return
                else:
                    logger.error("Unexpected track type: dict")
                    return

            file_list = self.downloader.get_file_list(tracks, voice_path)
            self.downloader.create_recover_file(file_list, voice_path)

        tasks = []
        for rj_id in ids:
            tasks.append(update_one(rj_id))

        await asyncio.gather(*tasks)


class ASMRPlayListManager(AsyncManager):
    def __init__(self, name: str, password: str, proxy: str, limit: int = 3):
        self.playlist = ASMRPlayListAPI(
            name=name, password=password, proxy=proxy, limit=limit
        )
        super().__init__(self.playlist)

    async def list(self, num: int = 12, raw: bool = False):
        from asmrmanager.common.output import print_table

        playlists, total = await self.playlist.get_playlists(
            page=1, page_size=num
        )
        print_table(
            titles=["id", "name", "amount", "privacy"],
            rows=[
                (str(p.id), p.name, p.works_count, p.privacy.name)
                for p in playlists
            ],
            raw=raw,
        )
        print(f"({len(playlists)}/{total})")

        fm.save_playlist_cache(playlists)

    async def remove(self, pl_ids: Iterable[uuid.UUID]):
        res = await asyncio.gather(*map(self.playlist.delete_playlist, pl_ids))

        if not isinstance(res, list):
            logger.error(f"Unexpected response type when delete playlists.")
            return
        for r in res:
            if not isinstance(r, dict):
                logger.error(
                    f"Unexpected response type when delete playlists."
                )
                return
            if r.get("error"):
                logger.error(f"Error when delete playlists: {r}")
                return
            logger.info(f"Sucessfully delete playlist {r['id']}.")

    async def create(
        self,
        name: str,
        desc: str | None,
        privacy: Literal["PUBLIC", "NON_PUBLIC", "PRIVATE"],
    ):
        from asmrmanager.spider.playlist import PRIVACY

        res = await self.playlist.create_playlist(name, desc, PRIVACY[privacy])
        if not isinstance(res, dict):
            logger.error(f"Unexpected response type when create playlist.")
            return
        if res.get("error"):
            logger.error(f"Error when creating playlist:{res}")
            return
        logger.info(f"Sucessfully create playlist: {res['id']}.")

    async def add(self, rj_ids: Iterable[RJID], pl_id: uuid.UUID):
        res = await self.playlist.add_works_to_playlist(rj_ids, pl_id)
        if not isinstance(res, dict):
            logger.error(
                f"Unexpected response type when add works to {pl_id}."
            )
            return
        if res.get("error"):
            logger.error(f"Error when add works to playlist:{res}")
            return
        logger.info(f"Sucessfully add works to playlist {pl_id}.")



# playlist manager
import uuid
from typing import List

import click
from typing_extensions import Literal

from asmrmanager.cli.core import (
    create_playlist,
    multi_rj_argument,
    pl_preprocess_cb,
)
from asmrmanager.common.rj_parse import RJID


@click.group()
def pl():
    """asmr playlist interface"""


@click.command("list")
@click.option(
    "--num",
    "-n",
    default=12,
    show_default=True,
    help="number of playlists to show",
)
@click.option(
    "--raw",
    "-r",
    is_flag=True,
    default=False,
    show_default=True,
    help="raw output",
)
def list_(num: int, raw: bool):
    """list all playlists"""
    pl = create_playlist()
    pl.run(pl.list(num, raw))


@click.command()
@multi_rj_argument
@click.argument("pl_id", callback=pl_preprocess_cb)
def add(rj_ids: List[RJID], pl_id: uuid.UUID):
    """add a playlist"""
    pl = create_playlist()
    pl.run(pl.add(rj_ids, pl_id))


@click.command("rm")
@click.argument("pl_ids", nargs=-1, callback=pl_preprocess_cb)
def remove(pl_ids: List[uuid.UUID]):
    """remove a playlist"""
    pl = create_playlist()
    pl.run(pl.remove(pl_ids))


@click.command()
@click.argument("name")
@click.option("--desc", "-d", default=None, help="playlist description")
@click.option(
    "--privacy",
    "-p",
    type=click.Choice(["PUBLIC", "NON_PUBLIC", "PRIVATE"]),
    default="PRIVATE",
)
def create(
    name: str,
    desc: str | None,
    privacy: Literal["PUBLIC", "NON_PUBLIC", "PRIVATE"],
):
    """create a playlist"""
    pl = create_playlist()
    pl.run(pl.create(name, desc, privacy))


pl.add_command(list_)
pl.add_command(add)
pl.add_command(remove)
pl.add_command(create)



import functools
import uuid
from typing import TYPE_CHECKING, List, Tuple

import click

from asmrmanager.common.browse_params import BrowseParams
from asmrmanager.common.download_params import DownloadParams
from asmrmanager.common.parse_filter import name_should_download
from asmrmanager.common.rj_parse import id2rj, rj2id
from asmrmanager.config import config
from asmrmanager.filemanager.manager import FileManager
from asmrmanager.logger import logger

fm = FileManager(config.storage_path, config.download_path, config.view_path)

if TYPE_CHECKING:
    from asmrmanager.database.manage import DataBaseManager
    from asmrmanager.spider import ASMRDownloadManager


def create_database():
    from asmrmanager.database.manage import DataBaseManager

    return DataBaseManager(tag_filter=config.tag_filter or tuple())


def create_downloader_and_database(
    download_params: DownloadParams | None = None,
) -> Tuple["ASMRDownloadManager", "DataBaseManager"]:
    from asmrmanager.spider import ASMRDownloadManager

    db = create_database()

    if download_params is None:
        download_params = DownloadParams(False, False, False)

    return (
        ASMRDownloadManager(
            name=config.username,
            password=config.password,
            proxy=config.proxy,
            id_should_download=(
                (lambda _: True)
                if download_params.force
                else (
                    lambda rj_id: (not db.check_exists(rj_id))
                    or (fm.get_location(rj_id) is None)
                )
            ),  # 如果数据库中不存在或者文件不存在，都执行下载
            json_should_download=db.add_info,
            name_should_download=(
                name_should_download
                if download_params.filter
                else (lambda *_: True)
            ),
            replace=download_params.replace,
            download_method=config.download_method,
            aria2_config=config.aria2_config,
        ),
        db,
    )


def create_playlist():
    from asmrmanager.spider.interface import ASMRPlayListManager

    return ASMRPlayListManager(
        name=config.username, password=config.password, proxy=config.proxy
    )


def browse_param_options(f):
    """pass the `browse_params` parameter to the function"""

    @click.option(
        "-p",
        "--page",
        type=int,
        default=1,
        help="page of the search result",
        show_default=True,
    )
    @click.option(
        "--subtitle/--no-subtitle",
        is_flag=True,
        default=False,
        help="if the ASMR has subtitle(中文字幕)",
        show_default=True,
    )  # no -s/-ns option for ocnfliction with --sell -s
    @click.option(
        "-o",
        "--order",
        type=click.Choice(
            [
                "create_date",
                "rating",
                "release",
                "dl_count",
                "price",
                "rate_average_2dp",
                "review_count",
                "id",
                "nsfw",
                "random",
            ],
            case_sensitive=False,
        ),
        default="release",
        help="ordering of the search result",
        show_default=True,
    )
    @click.option(
        "--asc/--desc", default=False, help="ascending or descending"
    )
    @functools.wraps(f)
    def wrapper_common_options(*args, **kwargs):
        keys = ["page", "subtitle", "order", "asc"]
        browse_params = BrowseParams(**{k: kwargs.pop(k) for k in keys})
        return f(*args, **kwargs, browse_params=browse_params)

    wrapper_common_options.__doc__ = (
        ""
        if wrapper_common_options.__doc__ is None
        else wrapper_common_options.__doc__
    )
    wrapper_common_options.__doc__ += """

    nsfw will only show the full age ASMRs

    for other --order values, you can refer to the website for
    explicit meaning
    """

    return wrapper_common_options


def download_param_options(f):
    """pass the `download_params` parameter to the function"""

    @click.option(
        "--force/--check-db",
        is_flag=True,
        default=False,
        help=(
            "force download even if the RJ id exists in database,"
            "or by default, RJ already in the database will be skipped"
        ),
    )
    @click.option(
        "--replace/--no-replace",
        is_flag=True,
        default=False,
        show_default=True,
        help="replace the file if it exists",
    )
    @click.option(
        "--filter/--no-filter",
        is_flag=True,
        default=True,
        show_default=True,
        help="filter out the files to download, rules are in the config file",
    )
    @functools.wraps(f)
    def wrap(*args, **kwargs):
        keys = ["force", "replace", "filter"]
        download_params = DownloadParams(**{k: kwargs.pop(k) for k in keys})
        return f(*args, **kwargs, download_params=download_params)

    wrap.__doc__ = "" if wrap.__doc__ is None else wrap.__doc__
    wrap.__doc__ += """

    --force will check the download RJ files again though it is already
     in the database, it work just like update

    --replace option will first delte the original file,
    then add the new file to download queue(i.e. IDM or aria2)
    """
    return wrap


PREVIOUS_RJ_PATH = fm.DATA_PATH / ".prev_rj"


def get_prev_rj():
    if not PREVIOUS_RJ_PATH.exists():
        return ""
    rj = PREVIOUS_RJ_PATH.read_text(encoding="utf8")
    logger.info(f"previous RJ id is {rj}")
    return rj


def save_rj(rj: str):
    PREVIOUS_RJ_PATH.write_text(rj, encoding="utf8")


def rj_argument(f):
    """parse the rj_id: int, if not given it will use the previous rj_id"""

    @click.argument("rj_id", type=str, default="__default__")
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        rj: str = kwargs["rj_id"]
        if kwargs["rj_id"] == "__default__":
            rj = get_prev_rj()
        if rj == "":
            logger.error(
                "No previous RJ id available,"
                " please first run a command with Rj id"
            )
            exit(-1)

        rj_id = rj2id(rj)
        if rj_id is None:
            logger.error(f"Invalid input RJ ID{rj}")
            exit(-1)
        save_rj(rj)
        kwargs["rj_id"] = rj_id

        f(*args, **kwargs)

    return wrapper


def multi_rj_argument(f):
    """parse multiple rj input to rj_id"""

    @click.argument("rjs", nargs=-1)
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        rjs: Tuple[str] = kwargs["rjs"]
        del kwargs["rjs"]
        kwargs["rj_ids"] = []
        for rj in rjs:
            rj_id = rj2id(rj)
            if rj_id is None:
                logger.error(f"Invalid input RJ ID{rj}")
                continue
            kwargs["rj_ids"].append(rj_id)

        if len(kwargs["rj_ids"]) == 0:
            rj = get_prev_rj()
            if rj == "":
                logger.error(
                    "No previous RJ id available,"
                    " please first run a command with Rj id"
                )
                exit(-1)

            rj_id = rj2id(rj)
            if rj_id is None:
                logger.error(f"Invalid input RJ ID{rj}")
                exit(-1)
            kwargs["rj_ids"].append(rj_id)
        elif len(kwargs["rj_ids"]) == 1:
            save_rj(id2rj(kwargs["rj_ids"][0]))

        f(*args, **kwargs)

    return wrapper


SEPARATOR = ":"


def interval_preprocess_cb(ctx: click.Context, opt: click.Parameter, val: str):
    """
    input must be numeric or None,
    and it returns a float/int/None value tuple
    """
    if val is None:
        return (None, None)
    vals = val.split(SEPARATOR)
    assert len(vals) == 2

    def _check(x: str):
        if x == "":
            return None
        try:
            xf: float = float(x)
        except ValueError:
            ctx.fail(f"{x} is not a valid number")

        assert isinstance(xf, float)
        if xf.is_integer():
            return int(xf)
        return xf

    return tuple(map(_check, vals))


def pl_preprocess_cb(
    ctx: click.Context, param: click.Option, val: str | Tuple
) -> List[uuid.UUID]:
    def is_valid_uuid(v: str):
        try:
            uuid.UUID(v)
            return True
        except ValueError:
            return False

    if isinstance(val, str):
        val = (val,)

    res = []

    # check for valid uuid
    possible_aliaes = []
    for v in val:
        if is_valid_uuid(v):
            res.append(v)
            continue
        else:
            possible_aliaes.append(v)

    # make transfromation from alias to uuid or name
    possible_names = []
    for alias in possible_aliaes:
        name_or_uuid = config.playlist_aliases.get(alias, alias)
        if is_valid_uuid(name_or_uuid):
            res.append(name_or_uuid)
        else:
            possible_names.append(name_or_uuid)

    # check and get cached playlist
    playlists_cache = fm.get_playlist_cache()
    if playlists_cache is None:
        logger.warning("no cached playlist, try getting from server")
        pl = create_playlist()
        # TODO use to CONSTANT to replace 100
        playlists_cache, total = pl.run(
            pl.playlist.get_playlists(page_size=100)
        )[0]
        if total > len(playlists_cache):
            logger.warning(
                f"total playlist number {total} is larger than"
                f" {len(playlists_cache)}"
            )
        fm.save_playlist_cache(playlists_cache)
    cached_names = [p.name for p in playlists_cache]

    # check for valid names using playlist cache
    for name in possible_names:
        if name in cached_names:
            res.append(playlists_cache[cached_names.index(name)].id)
        else:
            logger.warning(f"invalid playlist name {name}")

    if not res:
        logger.warning(
            "If you just have created a playlist, you may need to use pl list"
            " to cache the playlists to local"
        )
        ctx.fail("no valid uuid, alias or names found")

    if param.nargs == 1:
        return res[0]
    return res



import click

from asmrmanager.cli.core import rj_argument
from asmrmanager.common.rj_parse import RJID
from asmrmanager.logger import logger


@click.command()
@rj_argument
@click.option("-s", "--star", type=int, help="must be integer in [1, 5]")
@click.option("-c", "--comment", type=str)
def review(rj_id: RJID, star: int, comment: str):
    """review an ASMR with star(1-5) and comment and add it to storage path"""
    logger.info(
        f"run command review with rj_id={rj_id}, star={star} comment={comment}"
    )

    from asmrmanager.cli.core import create_database, fm

    db = create_database()

    update_stored = False
    if not fm.could_store():
        logger.warning("storage path not found skip storing operation")
    else:
        fm.store(rj_id)
        update_stored = True
    db.update_review(rj_id, star, comment, update_stored=update_stored)
    db.commit()



import logging
from logging.handlers import TimedRotatingFileHandler
from os import makedirs
from sys import stderr

from colorlog import ColoredFormatter

from asmrmanager.filemanager.appdirs_ import LOG_PATH

log_path = LOG_PATH
makedirs(log_path, exist_ok=True)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

console_formatter = ColoredFormatter(
    "%(asctime)s - %(log_color)s%(levelname)s%(reset)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    reset=True,
)

file_formatter = logging.Formatter(
    "%(asctime)s - %(filename)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
console_handler = logging.StreamHandler(stderr)
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(console_formatter)

file_handler = TimedRotatingFileHandler(
    log_path / "asmr.log",
    when="D",
    backupCount=1,
    encoding="utf8",
)
file_handler.setLevel(logging.DEBUG)
file_handler.setFormatter(file_formatter)

logger.addHandler(console_handler)
logger.addHandler(file_handler)



from pathlib import Path
from typing import List, Tuple

import click
from textual.app import App, ComposeResult
from textual.containers import Center
from textual.widgets import Footer, Label, ProgressBar

from asmrmanager.common.vtt2lrc import vtt2lrc
from asmrmanager.logger import logger

from .player import MusicPlayer, MusicPlayerWithLyrics

# from textual.reactive import reactive


class LRCPlayer(App):
    CSS_PATH = "main.css"
    BINDINGS = [
        ("space", "pause", "暂停"),
        ("j", "forward", "快进"),
        ("k", "backward", "快退"),
        ("l", "next_voice", "下一首"),
        ("h", "prev_voice", "上一首"),
    ]

    def __init__(self, episodes: List[Tuple], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.episodes = episodes
        self.players: List[MusicPlayer | None] = [None] * len(self.episodes)
        self.voice_index = 0

    @property
    def player(self) -> MusicPlayer | MusicPlayerWithLyrics:
        if self.players[self.voice_index] is None:
            if self.episodes[self.voice_index][1] is None:
                self.players[self.voice_index] = MusicPlayer(
                    self.episodes[self.voice_index][0]
                )  # pyright: ignore
            else:
                self.players[self.voice_index] = MusicPlayerWithLyrics(
                    *self.episodes[self.voice_index]
                )

        return self.players[self.voice_index]  # pyright: ignore

    def on_mount(self):
        self.player.play()
        self.progress_timer = self.set_interval(0.05, self.progress)

    def progress(self):
        if not self.player.is_playing():
            self.action_next_voice()
            return

        if isinstance(self.player, MusicPlayerWithLyrics):
            lrc_data = self.player.get_lyrics()
            self.query_one(ProgressBar).update(progress=lrc_data.progress)
            for i, l in enumerate(self.query(".lyrics")):
                assert isinstance(l, Label)
                l.update(lrc_data.lyrics[i])

        else:
            self.query_one(ProgressBar).update(
                progress=self.player.get_total_percent()
            )
            for i, l in enumerate(self.query(".lyrics")):
                assert isinstance(l, Label)
                l.update(("", "no lyrics", "")[i])

        # 更新播放时间
        # 格式化成 x:xx / x:xx
        def format_time(seconds: int):
            return f"{seconds // 60_000}:{seconds % 60_000 // 1000:02d}"

        self.query_one("#time", expect_type=Label).update(
            f"[{format_time(self.player.get_pos())}"
            " / "
            f"{format_time(self.player.total_time)}]"
        )

    def compose(self) -> ComposeResult:
        yield Label(self.player.title, id="title")

        for i in range(3):
            label = Label("", classes="lyrics")
            label.styles.text_opacity = str(100 - abs(1 - i) * 50) + "%"
            yield label

        with Center():
            yield ProgressBar(100, show_eta=False, id="pgbar")
            yield Label("[0:00 / 0:00]", id="time")

        yield Footer()

    def action_pause(self):
        if self.player.is_playing():
            self.player.pause()
            self.progress_timer.pause()
        else:
            self.player.unpause()
            self.progress_timer.resume()

    def action_forward(self, t: float | None = None):
        if isinstance(self.player, MusicPlayerWithLyrics):
            self.player.forward_lrc()
        else:
            self.player.forward(t)

    def action_backward(self, t: float | None = None):
        if isinstance(self.player, MusicPlayerWithLyrics):
            self.player.backward_lrc()
        else:
            self.player.backward(t)

    def switch_voice(self, idx: int):
        self.voice_index = idx
        self.player.play()
        self.query_one("#title", expect_type=Label).update(self.player.title)

    def action_next_voice(self):
        if self.voice_index + 1 < len(self.episodes):
            self.switch_voice(self.voice_index + 1)

    def action_prev_voice(self):
        if self.voice_index - 1 >= 0:
            self.switch_voice(self.voice_index - 1)


@click.command()
@click.argument(
    "path", type=click.Path(exists=True, dir_okay=True, path_type=Path)
)
def main(path: Path):
    import os

    # 每个元素是一个元组，包含了文件名和歌词文件名，如果没有歌词则为None
    episodes: List[Tuple[Path, Path | None]] = []
    if path.is_dir():
        for file in os.listdir(path):
            file_path = path / file
            if file_path.suffix in (".mp3", ".wav", ".m4a", ".flac"):
                lrc_path = file_path.with_suffix(".lrc")
                if not lrc_path.exists():
                    lrc_path = None
                episodes.append((file_path, lrc_path))
    else:
        if path.suffix in (".mp3", ".wav", ".m4a", ".flac"):
            lrc_path = path.with_suffix(".lrc")
            vtt_path = path.with_suffix(path.suffix + ".vtt")
            if (not lrc_path.exists()) and (not vtt_path.exists()):
                subtitle_path = None

            match (lrc_path.exists(), vtt_path.exists()):
                case (True, True) | (True, False):
                    subtitle_path = lrc_path
                case (False, True):
                    logger.info(f"convert vtt to lrc: {vtt_path}")
                    lrc_content = vtt2lrc(vtt_path)
                    with open(lrc_path, "w", encoding="utf-8") as f:
                        f.write(lrc_content)
                    subtitle_path = lrc_path
                case _:
                    subtitle_path = None
            episodes.append((path, subtitle_path))

    if not episodes:
        print("error input")
        return

    app = LRCPlayer(episodes)
    app.run()


if __name__ == "__main__":
    main()



from pathlib import Path

import appdirs

from asmrmanager import __name__ as package_name

CACHE_PATH = Path(appdirs.user_cache_dir(package_name))
CONFIG_PATH = Path(appdirs.user_config_dir(package_name))
DATA_PATH = Path(appdirs.user_data_dir(package_name))
LOG_PATH = Path(appdirs.user_log_dir(package_name))



import re
from pathlib import Path
from typing import Any, Dict, List, Tuple

import click

from asmrmanager.cli.core import (
    create_database,
    fm,
    multi_rj_argument,
    rj_argument,
)
from asmrmanager.common.rj_parse import RJID
from asmrmanager.filemanager.exceptions import DstItemAlreadyExistsException
from asmrmanager.logger import logger


@click.group()
def file():
    """file management"""


@click.command("del")
@rj_argument
def del_(rj_id: RJID):
    """not implemented"""
    raise NotImplementedError
    from asmrmanager.cli.core import create_fm
    from asmrmanager.filemanager.utils import folder_chooser_multiple

    fm = create_fm()

    rj_path = fm.get_path(rj_id)

    if rj_path is None:
        logger.error(f"item {rj_id} does not exists")
        return

    folders = folder_chooser_multiple(
        rj_path,
        lambda p: any(
            [i.suffix != ".info" for i in p.iterdir() if not i.is_dir()]
        ),
    )

    for folder in folders:
        pass


@click.command()
@rj_argument
@click.option(
    "--regex", "-r", type=str, default=".*", help="use regex to match"
)
@click.option(
    "--ignore-filter",
    "-i",
    "ignore_filter",
    is_flag=True,
    default=False,
    show_default=True,
    help="recover files that have been filtered out",
)
def recover(rj_id: RJID, regex: str, ignore_filter: bool):
    """recover a file from recover file"""
    rj_path = fm.get_path(rj_id)
    if rj_path is None:
        logger.error(f"item {rj_id} does not exist")
        return

    recover_path = rj_path / ".recover"
    if not recover_path.exists():
        logger.error(
            f"item {rj_id} does not have recover file, please update this"
            " rj id first"
        )
        return

    import json

    recovers: List[Dict[str, Any]] = json.loads(
        recover_path.read_text(encoding="utf8")
    )

    url2download: List[Tuple[str, Path]] = []

    for recover in recovers:
        rel_path = recover["path"]

        recover_file = rj_path / rel_path
        if recover_file.exists():
            continue

        if not re.search(regex, rel_path):
            continue

        if not recover["should_download"]:
            if not ignore_filter:
                logger.info(
                    f"file {rel_path} is filtered out and should not be"
                    " recovered"
                )
                continue
            else:
                logger.warning(
                    f"recover file {rel_path} since filters are ignored"
                )

        url2download.append((recover["url"], recover_file))

    from asmrmanager.cli.core import create_downloader_and_database

    downloader, _ = create_downloader_and_database()
    tasks = []
    for url, path in url2download:
        tasks.append(
            downloader.downloader.process_download(url, path.parent, path.name)
        )

    downloader.run(*tasks)


@click.command()
@multi_rj_argument
@click.option(
    "--replace/--no-replace",
    "-r/-nr",
    is_flag=True,
    default=True,
    show_default=True,
    help="replace the files if exists",
)
@click.option(
    "--all",
    "-a",
    "all_",
    is_flag=True,
    default=False,
    show_default=True,
    help="store all files",
)
def store(rj_ids: List[RJID], replace: bool, all_: bool):
    """
    store the downloaded files to the storage
    """

    db = create_database()
    try:
        if all_:
            import cutie

            res = cutie.prompt_yes_or_no(
                "Are you sure to store all files in the download_path?",
            )
            if res is None or res is False:
                return
            fm.store_all(replace=replace)
            id_to_store = fm.list_("download")

        else:
            for rj_id in rj_ids:
                fm.store(rj_id, replace=replace)
            id_to_store = rj_ids

        for id_ in id_to_store:
            res = db.check_exists(id_)
            if not res:
                logger.error(
                    "no such id: %s, which is an unexpected situation", id_
                )
                continue
            res.stored = True
        db.commit()
        logger.info("succesfully stored all files")
    except DstItemAlreadyExistsException as e:
        logger.error("storing terminated for %s", e)


file.add_command(del_)
file.add_command(recover)
file.add_command(store)



from typing import Optional

from sqlalchemy import func
from sqlalchemy.orm import Session

from asmrmanager.database.orm_type import ASMRInstance

from .database import ASMR, Tag


class QFunc:
    def __init__(self, ss: Session):
        self.ss = ss

    def get_info(self, rj_id: int, rand: bool) -> ASMRInstance | None:
        if rand:
            return self.ss.query(ASMR).order_by(func.random()).first()
        return self.ss.query(ASMR).get(rj_id)

    def get_tag_id(self, name: str) -> Optional[int]:
        res = self.ss.query(Tag).filter(Tag.name == name).one_or_none()
        if res:
            return res.id
        return None

    def get_tag_name(self, tid: int):
        res = self.ss.query(Tag).filter(Tag.id == tid).one_or_none()
        if res:
            return res.name
        return None

    def get_stored(self, rj_id: int):
        res = self.ss.query(ASMR).get(rj_id)
        return res.stored if res is not None else None



import asyncio
from typing import Any, Dict, List, TypeVar

from aiohttp import ClientConnectorError, ClientSession
from aiohttp.connector import TCPConnector

from asmrmanager.common.rj_parse import RJID, id2rj
from asmrmanager.logger import logger

T = TypeVar("T", bound="ASMRAPI")


class ASMRAPI:
    base_api_url = "https://api.asmr-200.com/api/"

    headers = {
        "Referer": "https://www.asmr.one/",
        # ---
        # 这些参数中的某一个或几个应该可以解决
        # 如RJ296187会出现的 504 Gateway Time-out 问题
        "Origin": "https://www.asmr.one",
        "Host": "api.asmr-200.com",
        "Connection": "keep-alive",
        "Sec-Fetch-Mode": "cors",
        "Sec-Fetch-Site": "cross-site",
        "Sec-Fetch-Dest": "empty",
        # ---
        "User-Agent": (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
            "AppleWebKit/537.36 (KHTML, like Gecko)"
            " Chrome/78.0.3904.108 Safari/537.36"
        ),
    }

    def __init__(
        self, name: str, password: str, proxy: str, limit: int
    ) -> None:
        self._session: ClientSession
        self.name = name
        self.password = password
        self.proxy = proxy
        self.limit = limit

    async def login(self) -> None:
        try:
            async with self._session.post(
                self.base_api_url + "auth/me",
                json={"name": self.name, "password": self.password},
                headers=self.headers,
                proxy=self.proxy,
            ) as resp:
                token = (await resp.json())["token"]
                self.headers.update(
                    {
                        "Authorization": f"Bearer {token}",
                    }
                )
        except ClientConnectorError as err:
            logger.error(f"Login failed, {err}")

    async def get(self, route: str, params: dict | None = None) -> Any:
        resp_json = None
        while not resp_json:
            try:
                async with self._session.get(
                    self.base_api_url + route,
                    headers=self.headers,
                    proxy=self.proxy,
                    params=params,
                ) as resp:
                    resp_json = await resp.json()
                    return resp_json
            except Exception as e:
                logger.warning(f"Request {route} failed: {e}")
                await asyncio.sleep(3)
        return resp_json

    async def post(self, route: str, data: dict | None = None) -> Any:
        resp_json = None
        while not resp_json:
            try:
                async with self._session.post(
                    self.base_api_url + route,
                    headers=self.headers,
                    proxy=self.proxy,
                    json=data,
                ) as resp:
                    resp_json = await resp.json()
                    return resp_json
            except Exception as e:
                logger.warning(f"Request {route} failed: {e}")
                await asyncio.sleep(3)
        return resp_json

    async def get_playlists(
        self, page, page_size: int = 12, filter_by: str = "all"
    ) -> Dict[str, Any]:
        return await self.get(
            "playlist/get-playlists",
            params={
                "page": page,
                "pageSize": page_size,
                "filterBy": filter_by,
            },
        )

    async def create_playlist(
        self, name: str, desc: str | None = None, privacy: int = 0
    ):
        # Literal['public', 'non-public', 'private']
        return await self.post(
            "playlist/create-playlist",
            data={
                "name": name,
                "description": desc or "",
                "privacy": privacy,
                "locale": "zh-CN",
                "works": [],
            },
        )

    async def add_works_to_playlist(self, rj_ids: List[RJID], pl_id: str):
        return await self.post(
            "playlist/add-works-to-playlist",
            data={"id": pl_id, "works": rj_ids},
        )

    async def delete_playlist(self, pl_id: str):
        return await self.post("playlist/delete-playlist", data={"id": pl_id})

    async def get_search_result(
        self, content: str, params: dict
    ) -> Dict[str, Any]:
        return await self.get(f"search/{content}", params=params)

    async def list(self, params: dict) -> Dict[str, Any]:
        return await self.get("works", params=params)

    async def tag(self, tag_name: str, params: dict):
        # return await self.get(f'tags/{tag_id}/works', params=params)
        return await self.get_search_result(f"$tag:{tag_name}$", params=params)

    async def va(self, va_name: str, params: dict):
        return await self.get_search_result(f"$va:{va_name}$", params=params)

    async def __aenter__(self: T) -> T:
        self._session = ClientSession(connector=TCPConnector(limit=self.limit))
        await self.login()
        return self

    async def __aexit__(self, *_) -> None:
        await self._session.close()



import uuid
from dataclasses import dataclass
from enum import Enum
from typing import NewType

RJID = NewType("RJID", int)
RJName = NewType("RJName", str)


class PRIVACY(Enum):
    PUBLIC = 2
    NON_PUBLIC = 1
    PRIVATE = 0


@dataclass
class PlayListItem:
    id: uuid.UUID
    name: str
    privacy: PRIVACY
    desc: str
    works_count: int
    latest_work_id: RJID | None = None

    def asdict(self):
        return {
            "id": str(self.id),
            "name": self.name,
            "privacy": self.privacy.value,
            "desc": self.desc,
            "works_count": self.works_count,
            "latest_work_id": self.latest_work_id,
        }



from asmrmanager.filemanager.appdirs_ import CONFIG_PATH, DATA_PATH
from asmrmanager.filemanager.manager import FileManager

if not (DATA_PATH / "sqls").exists():
    FileManager.init_sqls()

if not (CONFIG_PATH / "config.toml").exists():
    FileManager.init_config()



import uuid

from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.types import CHAR, TypeDecorator


class GUID(TypeDecorator):
    """Platform-independent GUID type.
    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.
    """

    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def _uuid_value(self, value):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value

    def process_result_value(self, value, dialect):
        return self._uuid_value(value)

    def sort_key_function(self, value):
        return self._uuid_value(value)



from asmrmanager.cli.core import fm

_instance = None


def get_engine(check_same_thread: bool = False):
    from sqlalchemy import create_engine

    if _instance is not None:
        return _instance

    db_path = fm.DATA_PATH / "data.db"
    return create_engine(
        f"sqlite:///{db_path}?check_same_thread={check_same_thread}"
    )



import json
from pathlib import Path

import aioaria2

from asmrmanager.logger import logger


class Aria2Downloader:
    def __init__(
        self,
        proxy: str,
    ) -> None:
        self.client: aioaria2.Aria2HttpClient
        # logger.info(f"Connecting to aria2 rpc server: {self.client}")
        # self.api = aria2p.API(self.client)
        # self.options = aria2p.Options(
        #     self.api,
        #     {"all-proxy": proxy}
        # )
        # self.options.auto_file_renaming = False
        self.options = {
            "all-proxy": proxy,
            "auto-file-renaming": False,
        }

    async def create_client(self, host: str, port: int, secret: str) -> None:
        self.client = await aioaria2.Aria2HttpClient(
            f"{host}:{port}/jsonrpc",
            token=secret,
            loads=json.loads,
            dumps=json.dumps,
        ).__aenter__()

    async def close_client(self) -> None:
        await self.client.__aexit__(None, None, None)

    async def download(self, url: str, save_path: Path, filename: str):
        download_option = {
            "out": filename,
            "dir": str(save_path),
        }
        await self.client.addUri(
            [url], options={**self.options, **download_option}
        )



import os
import shutil
from pathlib import Path
from typing import Iterable, List, Literal, NamedTuple

import toml

from asmrmanager.common.rj_parse import RJID, id2rj, rj2id
from asmrmanager.common.types import PlayListItem
from asmrmanager.filemanager.appdirs_ import (
    CACHE_PATH,
    CONFIG_PATH,
    DATA_PATH,
    LOG_PATH,
)
from asmrmanager.filemanager.file_zipper import zip_chosen_folder
from asmrmanager.logger import logger

# from .exceptions import DstItemAlreadyExistsException, SrcNotExistsException


class FileManager:
    CONFIG_PATH = CONFIG_PATH
    DATA_PATH = DATA_PATH
    LOG_PATH = LOG_PATH
    CACHE_PATH = CACHE_PATH

    @classmethod
    def init_config(cls):
        cls.CONFIG_PATH.mkdir(parents=True, exist_ok=True)
        dst_path = cls.CONFIG_PATH / "config.toml"
        if dst_path.exists():
            return

        shutil.copy(
            Path(__file__).parent / "resources" / "config.example.toml",
            dst_path,
        )
        logger.info(
            f"An example config file has been copied to {dst_path}, please"
            " modify it and run this command again"
        )
        exit(0)

    @classmethod
    def init_sqls(cls):
        cls.DATA_PATH.mkdir(parents=True, exist_ok=True)
        dst_path = cls.DATA_PATH / "sqls"
        if dst_path.exists():
            return

        shutil.copytree(
            Path(__file__).parent / "resources" / "sqls.example",
            dst_path,
        )
        logger.info(f"First time to run, copy default sqls to {dst_path}")

    @classmethod
    def get_playlist_cache(cls) -> List[PlayListItem] | None:
        cls.CACHE_PATH.mkdir(parents=True, exist_ok=True)
        dst_path = cls.CACHE_PATH / "playlist.cache"
        if dst_path.exists():
            playlists = toml.load(dst_path)["playlists"]
            return [PlayListItem(**p) for p in playlists]

    @classmethod
    def save_playlist_cache(cls, playlists: List[PlayListItem]):
        cls.CACHE_PATH.mkdir(parents=True, exist_ok=True)
        dst_path = cls.CACHE_PATH / "playlist.cache"
        toml.dump(
            {"playlists": [p.asdict() for p in playlists]},
            dst_path.open(mode="w", encoding="utf-8"),
        )

    def __init__(
        self,
        storage_path: str,
        download_path: str,
        view_path: str,
    ):
        self.storage_path = Path(storage_path)
        self.download_path = Path(download_path)
        self.view_path = Path(view_path)

        self.storage_path_exists = (
            True if os.path.exists(self.storage_path) else False
        )
        self.download_path_exists = (
            True if os.path.exists(self.download_path) else False
        )
        self.view_path_exists = (
            True if os.path.exists(self.view_path) else False
        )

    def could_store(self):
        return self.storage_path_exists and self.download_path_exists

    def store(self, rj_id: RJID, replace=True):
        """sync download path and storage path"""
        # TODO 换用fcp实现？
        assert self.could_store()
        rj_name = id2rj(rj_id)
        if not os.path.exists(self.download_path / rj_name):
            logger.warning(f"item {rj_name} does not exists, skip it")
            return

        # if os.path.exists(self.storage_path / rj_name):
        #     if not exists_ok:
        #         logger.error(f"the item {rj_name} to store already exists!")
        #         raise DstItemAlreadyExistsException

        #     logger.info(f"remove item {rj_name} in storage")
        #     shutil.rmtree(self.storage_path / rj_name)

        logger.info(
            f"move {self.download_path / rj_name} "
            f"to {self.storage_path / rj_name}"
        )

        for root, _, files in os.walk(self.download_path / rj_name):
            dst_root = self.storage_path / Path(root).relative_to(
                self.download_path
            )
            dst_root.mkdir(parents=True, exist_ok=True)
            for file in files:
                src_file = Path(root) / file
                dst_file = dst_root / file

                if dst_file.exists():
                    if not replace:
                        logger.info(f"file {dst_file} already exists, skip it")
                        continue
                    logger.info(f"In replace mode, remove {dst_file}")
                    os.remove(dst_file)

                logger.info(f"move {src_file} to {dst_file}")
                shutil.move(src_file, dst_file)

        shutil.rmtree(self.download_path / rj_name)

        # try:
        #     shutil.copytree(
        #         self.download_path / rj_name,
        #         self.storage_path / rj_name,
        #         copy_function=shutil.copy2,
        #     )
        # except (FileNotFoundError, PermissionError, shutil.Error) as e:
        #     logger.error(
        #         f"moving files error: {e}, terminated, please munually clean"
        #         f" the files at {self.storage_path / rj_name}"
        #     )
        #     exit(-1)

        # try:
        #     shutil.rmtree(self.download_path / rj_name)
        # except (FileNotFoundError, PermissionError, shutil.Error) as e:
        #     logger.error(
        #         f"deleting original files error: {e}, terminated, please"
        #         f" munually clean the files at {self.download_path / rj_name}"
        #     )
        #     exit(-1)

    def store_all(self, replace=True):
        for file in os.listdir(self.download_path):
            rj_id = rj2id(file)
            if rj_id is None:
                logger.warning(f"Ignore invalid file {file} in download path")
                continue
            self.store(rj_id, replace=replace)

    def could_view(self):
        """
        either storage_path or download_path exists ok,
        view path must exists
        """
        return self.view_path_exists and (
            self.storage_path_exists or self.download_path_exists
        )

    @staticmethod
    def link(src: Path, dst: Path, hardlink=False):
        if hardlink:
            os.link(src, dst)
        else:
            os.symlink(src, dst)

    @staticmethod
    def _copy(src: Path, dst: Path, depth: int = -1):
        if src.is_file():
            shutil.copy(src, dst)
            return

        if depth == 0:
            return

        dst.mkdir(parents=True, exist_ok=True)

        for subfile in src.iterdir():
            FileManager._copy(subfile, dst / subfile.name, depth - 1)

    def remove_view(self, rj_id: RJID):
        assert self.could_view()
        rj_name = id2rj(rj_id)

        for path in self.view_path.iterdir():
            if path.stem != rj_name:
                continue
            if path.is_dir() and path.is_symlink():
                os.remove(path)
                return
            elif path.suffix == ".zip":
                os.remove(path)
                return
        else:
            logger.error(f"file {rj_name} not exists")
            return
        # path = self.view_path / rj_name
        # assert path.is_dir() and path.is_symlink()
        # os.remove(path)

    def list_(
        self, path: Literal["download", "view", "storage"]
    ) -> Iterable[RJID]:
        if path == "download":
            p = self.download_path
        elif path == "view":
            p = self.view_path
        elif path == "storage":
            p = self.storage_path
        else:
            logger.error("Invalid path")
            return []

        for name in p.iterdir():
            rj_id = rj2id(name.name)
            if rj_id is None:
                continue
            yield rj_id

    def zip_file(self, src: Path, dst: Path):
        zip_chosen_folder(src, dst)

    def get_location(
        self, rj_id: RJID
    ) -> Literal["download", "storage", None]:
        rj_name = id2rj(rj_id)
        if (self.storage_path / rj_name).exists():
            return "storage"
        if (self.download_path / rj_name).exists():
            return "download"
        return None

    def get_path(self, rj_id: RJID) -> Path | None:
        """get rj file path"""
        res = self.get_location(rj_id)

        match res:
            case "download":
                return self.download_path / id2rj(rj_id)
            case "storage":
                return self.storage_path / id2rj(rj_id)
            case _:
                return None

    def check_exists(self, rel_path: str):
        download, storage = False, False
        download_file_path = self.download_path / rel_path
        storage_file_path = self.storage_path / rel_path
        if download_file_path.exists():
            download = True
        if storage_file_path.exists():
            storage = True
        if (not download) and self.check_wav_flac_duplicate(
            download_file_path
        ):
            download = True
        if (not storage) and self.check_wav_flac_duplicate(storage_file_path):
            storage = True

        return NamedTuple(
            "ExistInfo", [("download", bool), ("storage", bool)]
        )(download, storage)

    @staticmethod
    def check_wav_flac_duplicate(file_path: Path) -> bool:
        """if file duplicate or already exists, return True"""
        if file_path.exists():
            logger.error(
                f"file already exists: {file_path}, please "
                "check for the existence of the download files first"
            )
            return True
        match file_path.suffix.lower():
            case ".wav":
                another_file_path = file_path.with_suffix(".flac")
                if another_file_path.exists():
                    logger.info(f"Detected {file_path} for same flac exists")
                    return True
            case ".flac":
                another_file_path = file_path.with_suffix(".wav")
                if another_file_path.exists():
                    logger.info(f"Detected {file_path} for same wav exists")
                    return True
            case _:
                pass

        return False



from datetime import date
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from asmrmanager.common.rj_parse import RJID


class ASMRInstance:
    id: "RJID"
    title: str
    circle_name: str
    tags: list["TagInstance"]
    vas: list["VoiceActorInstance"]
    nsfw: bool
    release_date: date
    price: int
    dl_count: int
    has_subtitle: bool

    star: int
    count: int
    comment: str

    held: bool
    stored: bool


class TagInstance:
    id: int
    name: str
    jp_name: str
    cn_name: str
    en_name: str


class VoiceActorInstance:
    id: int
    name: str



from typing import Literal


class BrowseParams:
    def __init__(
        self,
        page: int,
        subtitle: bool,
        order: Literal[
            "create_date",
            "rating",
            "release",
            "dl_count",
            "price",
            "rate_average_2dp",
            "review_count",
            "id",
            "nsfw",
            "random",
        ],
        asc: bool,
    ):
        self.page = page
        self.subtitle = subtitle
        self.order = order
        self.sort = "asc" if asc else "desc"

    @property
    def params(self):
        return {
            "page": str(self.page),
            "subtitle": str(int(self.subtitle)),
            "order": self.order,
            "sort": self.sort,
        }

    def __str__(self):
        return str(self.params)



import os
from collections import Counter
from pathlib import Path
from typing import Callable, List, Tuple

import cutie


def folder_chooser(
    folder: Path, path_filter: Callable[[Path], bool] = lambda _: True
) -> Path:
    assert folder.is_dir()
    choices: List[Tuple[Path, str]] = []
    for root, _, files in os.walk(folder):
        res = dict(Counter([Path(f).suffix for f in files]))
        desc = " , ".join([f"{k}: {v}" for k, v in res.items()])
        if path_filter(Path(root)):
            choices.append((Path(root), desc))
    index = cutie.select([f"{p} ({d})" for p, d in choices])
    return choices[index][0]


def folder_chooser_multiple(
    folder: Path, path_filter: Callable[[Path], bool] = lambda _: True
) -> List[Path]:
    assert folder.is_dir()
    choices: List[Tuple[Path, str]] = []
    for root, _, files in os.walk(folder):
        res = dict(Counter([Path(f).suffix for f in files]))
        desc = " , ".join([f"{k}: {v}" for k, v in res.items()])
        if path_filter(Path(root)):
            choices.append((Path(root), desc))
    indexes = cutie.select_multiple([f"{p} ({d})" for p, d in choices])
    return [choices[i][0] for i in indexes]


if __name__ == "__main__":
    print(folder_chooser(Path("/home/quy/sage")))
